/*
Réalisé par Denis Duplan pour Stash of Code (stashofcode@gmail.com) en mai 2021.

Cette oeuvre est mise à disposition selon les termes de la Licence Creative Commons Attribution - Pas d’Utilisation Commerciale 4.0 International (http://creativecommons.org/licenses/by-nc/4.0/).
*/

var libWEBGL = {

	// Fabriques

	createVSHelper: function (gl) {
		return (new this.VertexShaderHelper (gl));
	},
	createFSHelper: function (gl) {
		return (new this.FragmentShaderHelper (gl));
	},
	createMatrix2D: function (m) {
		return (new this.Matrix2D (m));
	},
	createVector2D: function (v) {
		return (new this.Vector2D (v));
	},
	createMatrix3D: function (m) {
		return (new this.Matrix3D (m));
	},
	createVector3D: function (v) {
		return (new this.Vector3D (v));
	}
};

/*******************************************************************************
Helpers pour vertex shader et fragment shader

Il s'agit de deux helpers pour faciliter la vie durant l'élaboration d'un shader. En effet, lors d'une telle élaboration, on revient souvent sur le format de point. Or cela implique de modifier de nombreux paramètres dans des appels à des fonctions de WebGL pour ce qui concerne les attributs utilisés par un vertex shader.

Il y a deux helpers ici : un pour un vertex shader, un pour un fragment shader.

Chaque objet s'utilise ainsi. Par exemple, pour un vertex shader :

	vsHelper = libWEBGL.createVSHelper (gl);
	vsHelper.addInput ("A_xy", libWEBGL.PRECISION_LOWP, libWEBGL.TYPE_VEC2);
	vsHelper.addInput ("A_mL", libWEBGL.PRECISION_LOWP, libWEBGL.TYPE_MAT3);
	vsHelper.addInput ("A_rgba", libWEBGL.PRECISION_LOWP, libWEBGL.TYPE_VEC4);
	vsHelper.addUniform ("U_mW", libWEBGL.PRECISION_LOWP, libWEBGL.TYPE_MAT3);
	vsHelper.addOutput ("V_rgba", libWEBGL.PRECISION_LOWP, libWEBGL.TYPE_VEC4);

Attention, car l'ordre importe : c'est le format de points qu'on décrit en ajoutant au fur et à mesure des inputs.

Une fois les variables inputs (attributs), les outputs (varying) et les uniformes spécifiées, il faut fournir le code GLSL du shader. Il sera automatiquement préfixé pour faire référence à toutes ces variables, un vertex shader sera créé sur cette base, et ce vertex shader sera compilé :

	shader = vsHelper.build ("void main (void) { gl_Position = vec4 (U_mW * A_mL * vec3 (A_xy.xy, 1.0), 1.0); V_rgba = A_rgba; }");

Il faut ensuite créer un programme et l'associer au shader helper. Le shader helper attache le shader au programme :

	program = gl.createProgram ();
	vsHelper.setProgram (program);

Enfin, il faut utiliser le shader helper, ce qui lui permet de recupèrer les locations de toutes variables du shader ajoutées au début, afin de pouvoir notamment configurer les attributs pour que le moteur de WebGL puisse lire les points dans le bon format :

	vsHelper.use ();

Ce dernier appel peut être effectué dans le contexte d'un VAO :

	vertexArray = gl.createVertexArray ();
	gl.bindVertexArray (vertexArray);
		gl.bindBuffer (gl.ELEMENT_ARRAY_BUFFER, bufferI);
		gl.bindBuffer (gl.ARRAY_BUFFER, bufferV);
		vsHelper.use ();
	gl.bindVertexArray (null);

Par la suite, il est possible de récupérer une référence sur les uniformes pour les modifier :

	gl.uniformMatrix3fv (vsHelper.getUniformLocation ("U_mW"), false, new Float32Array (getWorldMatrix ()));

A la fin du programme, il convient de détacher le shader du programme et de le supprimer, ainsi que de désactiver les attributs :

	vsHelper.release ();

En cours de route, il est possible de réinitialiser le helper, ce qui entre autres appelle la méthode précédente. Il est sans doute recommandé de d'avoir désactivé le programme qui utilise le vertex shader avant :

	gl.useProgram (null);
	gl.deleteProgram (program);
	vsHelper.reset ();

C'est la même chose pour le helper pour fragment shader, sauf que sa méthode .use () n'a pas à être appelée dans le contexte d'un VAO, puisqu'un tel shader n'utilise pas d'attributs.

Une fois ce code mis en place, toute modification du format de point nécessite simplement de modifier la liste des appels à .addInput (), .addOutput () et .addUniform () pour modifier la liste des variables utilisées par le code GLSL, et bien entendu de modifier l'appel à .build () avec le nouveau code GLSL du shader.
*******************************************************************************/

/*------------------------------------------------------------------------------
Fondation pour un helper
------------------------------------------------------------------------------*/

libWEBGL.ShaderHelper = function (gl, glType) {
	this.gl = gl;
	this.glType = glType;
	this.error = "";
	this.reset ();
};

libWEBGL.PRECISION_LOWP = 0;
libWEBGL.PRECISION_MEDIUMP = 1;
libWEBGL.PRECISION_HIGHP = 2;
libWEBGL.TYPE_FLOAT = 0;
libWEBGL.TYPE_VEC2 = 1;
libWEBGL.TYPE_VEC3 = 2;
libWEBGL.TYPE_VEC4 = 3;
libWEBGL.TYPE_MAT3 = 4;
libWEBGL.TYPE_MAT4 = 5;

libWEBGL.ShaderHelper.prototype.reset = function () {
	this.inputs = new Array ();		// Pour un VertexShader, des attributs (attribute) ; pour un fragment shader, des variantes (varying)
	this.outputs = new Array ();	// Pour un VertexShader, des variantes (varying) ; pour un fragment shader, l'équivalent de gl_FragColor
	this.uniforms = new Array ();
	this.release ();
};

libWEBGL.ShaderHelper.prototype.addInput = function (name, precision, type) {
	this.inputs.push ({ name: name, precision: precision, type: type, location: -1 });
};

libWEBGL.ShaderHelper.prototype.addOutput = function (name, precision, type) {
	this.outputs.push ({ name: name, precision: precision, type: type, location: -1 });
};

libWEBGL.ShaderHelper.prototype.addUniform = function (name, precision, type) {
	this.uniforms.push ({ name: name, precision: precision, type: type, location: -1 });
};

libWEBGL.ShaderHelper.prototype.setProgram = function (program) {
	this.program = program;
	this.gl.attachShader (this.program, this.shader);
};

libWEBGL.ShaderHelper.prototype.build = function (main) {
	var source, i, precisions, types;

	precisions = new Array ();
	precisions[libWEBGL.PRECISION_LOWP] = "lowp";
	precisions[libWEBGL.PRECISION_MEDIUMP] = "mediump";
	precisions[libWEBGL.PRECISION_HIGHP] = "highp";
	types = new Array ();
	types[libWEBGL.TYPE_FLOAT] = "float";
	types[libWEBGL.TYPE_VEC2] = "vec2";
	types[libWEBGL.TYPE_VEC3] = "vec3";
	types[libWEBGL.TYPE_VEC4] = "vec4";
	types[libWEBGL.TYPE_MAT3] = "mat3";
	types[libWEBGL.TYPE_MAT4] = "mat4";
	source = "#version 300 es\n";
	for (i = 0; i !== this.inputs.length; i ++)
		source += `in ${precisions[this.inputs[i].precision]} ${types[this.inputs[i].type]} ${this.inputs[i].name};`;
	for (i = 0; i !== this.uniforms.length; i ++)
		source += `uniform ${precisions[this.uniforms[i].precision]} ${types[this.uniforms[i].type]} ${this.uniforms[i].name};`;
	for (i = 0; i !== this.outputs.length; i ++)
		source += `out ${precisions[this.outputs[i].precision]} ${types[this.outputs[i].type]} ${this.outputs[i].name};`;
	source += main;
	this.shader = this.gl.createShader (this.glType);
	this.gl.shaderSource (this.shader, source);
	this.gl.compileShader (this.shader);
	if (!this.gl.getShaderParameter (this.shader, this.gl.COMPILE_STATUS)) {
		this.error = this.gl.getShaderInfoLog (this.shader);
		return (null);
	}
	return (this.shader);
};

libWEBGL.ShaderHelper.prototype.getError = function () {
	return (this.error);
};

libWEBGL.ShaderHelper.prototype.use = function () {
	var i;

	for (i = 0; i !== this.uniforms.length; i ++)
		this.uniforms[i].location = this.gl.getUniformLocation (this.program, this.uniforms[i].name);
};

libWEBGL.ShaderHelper.prototype.getUniformLocation = function (name) {
	var i;

	for (i = 0; i !== this.uniforms.length; i ++) {
		if (this.uniforms[i].name === name) {
			if (this.uniforms[i].location === -1)
				console.log ("libWEBGL warning: Uniform location is -1. Check if you must call .getUniformLocation () after .use ().");
			return (this.uniforms[i].location);
		}
	}
	return (-1);
};

libWEBGL.ShaderHelper.prototype.release = function () {
	if (this.program) {
		this.gl.detachShader (this.program, this.shader);
		this.program = null;
	}
	if (this.shader) {
		this.gl.deleteShader (this.shader);
		this.shader = null;
	}
};

/*------------------------------------------------------------------------------
Helper pour fragment shader
------------------------------------------------------------------------------*/

libWEBGL.FragmentShaderHelper = function (gl) {
	libWEBGL.ShaderHelper.call (this, gl, gl.FRAGMENT_SHADER);
};
libWEBGL.FragmentShaderHelper.prototype = Object.create (libWEBGL.ShaderHelper.prototype);

/*------------------------------------------------------------------------------
Helper pour vertex shader
------------------------------------------------------------------------------*/

libWEBGL.VertexShaderHelper = function (gl) {
	libWEBGL.ShaderHelper.call (this, gl, gl.VERTEX_SHADER);
};
libWEBGL.VertexShaderHelper.prototype = Object.create (libWEBGL.ShaderHelper.prototype);

libWEBGL.VertexShaderHelper.prototype.use = function () {
	var i, j, types, stride, offset;

	libWEBGL.ShaderHelper.prototype.use.call (this);
	for (i = 0; i !== this.inputs.length; i ++)
		this.inputs[i].location = this.gl.getAttribLocation (this.program, this.inputs[i].name);
	types = new Array ();
	types[libWEBGL.TYPE_FLOAT] = {size: 1, stride: 1};
	types[libWEBGL.TYPE_VEC2] = {size: 2, stride: 2};
	types[libWEBGL.TYPE_VEC3] = {size: 3, stride: 3};
	types[libWEBGL.TYPE_VEC4] = {size: 4, stride: 4};
	types[libWEBGL.TYPE_MAT3] = {size: 3, stride: 9};
	types[libWEBGL.TYPE_MAT4] = {size: 4, stride: 16};
	stride = 0;
	for (i = 0; i !== this.inputs.length; i ++)
		stride += types[this.inputs[i].type].stride;
	offset = 0;
	for (i = 0; i !== this.inputs.length; i ++) {
		switch (this.inputs[i].type) {
			case libWEBGL.TYPE_FLOAT:
			case libWEBGL.TYPE_VEC2:
			case libWEBGL.TYPE_VEC3:
			case libWEBGL.TYPE_VEC4:
				this.gl.vertexAttribPointer (this.inputs[i].location, types[this.inputs[i].type].size, this.gl.FLOAT, false, stride * Float32Array.BYTES_PER_ELEMENT, offset * Float32Array.BYTES_PER_ELEMENT);
				this.gl.enableVertexAttribArray (this.inputs[i].location);
				offset += types[this.inputs[i].type].size;
				break;
			case libWEBGL.TYPE_MAT3:
				for (j = 0; j !== 3; j ++) {
					this.gl.vertexAttribPointer (this.inputs[i].location + j, types[this.inputs[i].type].size, this.gl.FLOAT, false, stride * Float32Array.BYTES_PER_ELEMENT, offset * Float32Array.BYTES_PER_ELEMENT);
					this.gl.enableVertexAttribArray (this.inputs[i].location + j);
					offset += types[this.inputs[i].type].size;
				}
				break;
			case libWEBGL.TYPE_MAT4:
				for (j = 0; j !== 4; j ++) {
					this.gl.vertexAttribPointer (this.inputs[i].location + j, types[this.inputs[i].type].size, this.gl.FLOAT, false, stride * Float32Array.BYTES_PER_ELEMENT, offset * Float32Array.BYTES_PER_ELEMENT);
					this.gl.enableVertexAttribArray (this.inputs[i].location + j);
					offset += types[this.inputs[i].type].size;
				}
				break;
		}
	}
};

libWEBGL.VertexShaderHelper.prototype.release = function () {
	var i;

	for (i = 0; i !== this.inputs.length; i ++) {
		switch (this.inputs[i].type) {
			case libWEBGL.TYPE_FLOAT:
			case libWEBGL.TYPE_VEC2:
			case libWEBGL.TYPE_VEC3:
			case libWEBGL.TYPE_VEC4:
				this.gl.disableVertexAttribArray (this.inputs[i].location);
				break;
			case libWEBGL.TYPE_MAT3:
				for (j = 0; j !== 3; j ++)
					this.gl.disableVertexAttribArray (this.inputs[i].location + j);
				break;
			case libWEBGL.TYPE_MAT4:
				for (j = 0; j !== 4; j ++)
					this.gl.disableVertexAttribArray (this.inputs[i].location + j);
				break;
		}
	}
	libWEBGL.ShaderHelper.prototype.release.call (this);
};

/*******************************************************************************
Manipulation de matrices et de vecteurs * ATTENTION A L'OPTIMISATION SUR LES MULTIPLICATIONS ! LIRE CE QUI SUIT ! *

Contrairement à l'intuition, WebGL travaille avec des vecteurs et des matrices en colonne. Les élements d'une matrice sont numérotés en spécifiant d'abord le numéro de la colonne, puis le numéro de la ligne :

	m21 = colonne 2, ligne 1

Un vecteur étant une matrice sur une colonne, pour appliquer les transformations décrites par une matrice m à un vecteur v en GLSL, on écrit m puis v :

	v = m * v

Une conséquence de cette notation, c'est que pour transformer v par mA, puis par mB en GLSL, on écrit mB, puis mA, et enfin v :

	v = mB * mA * v

Par ailleurs, pour charger une matrice dans une uniforme par un appel à uniformMatrix3fv () en JavaScript, il faut fournir cette matrice sous la forme d'un tableau JavaScript où les éléments figurent colonne après colonne. Ainsi, si on considère qu'en GLSL, v = m * v correspond à :

					v0
					v1
					v2

	m00 m10 m20		v0 * m00 + v1 * m10 + v2 * m20
	m01 m11 m21		v1 * m01 + v1 * m11 + v2 * m21
	m02 m12 m22		v2 * m02 + v1 * m12 + v2 * m22

...alors il faut alimenter uniformMatrix3fv () avec :

	var m = [
		m00, m01, m02,
		m10, m11, m12,
		m20, m21, m22
	];

Comme en JavaScript, un vecteur correspond à une matrice sur une ligne, et non sur une colonne, la même multiplication s'écrit dans l'ordre inverse v = v * m, et implique la transposée de m :

				m00 m01 m02
				m10 m11 m12
				m20 m21 m22
	v0 v1 v2	v0 * m00 + v1 * m10 + v2 * m20
					v1 * m01 + v1 * m11 + v2 * m21
						v2 * m02 + v1 * m12 + v2 * m22

Le fait que les éléments d'une matrice soient stockés colonne après colonne dans un tableau JavaScript présente un intérêt, qui est de permettre d'écrire une multiplication en ligne en JavaScript sans avoir à transposer la matrice :

	               *------>| 0 1 2 |
	               | *---->| 3 4 5 | Multiplication en colonne (GLSL)
	               | | *-->| 6 7 8 |
	               | | |          
	     *------>| 0 1 2 |
	     | *---->| 3 4 5 |
	     | | *-->| 6 7 8 |
	     | | |
	   | 0 1 2 |
	   | 3 4 5 | Multiplication en ligne (JavaScript)
       | 6 7 8 |

Ce qui donne :

	var v = [v0, v1, v2];
	var m = [
		m00, m01, m02,
		m10, m11, m12,
		m20, m21, m22
	];
	
	v[0] = v[0] * m[0] + v[1] * m[3] + v[2] * m [6];
	v[1] = v[0] * m[1] + v[1] * m[4] + v[2] * m [7];
	v[2] = v[0] * m[2] + v[1] * m[5] + v[2] * m [8];

Cela permet de manipuler les matrices en JavaScript, et de fournir directement le résultat à WebGL. Par contre, il en résulte qu'écrire mB * mA * v en GLSL - transformer v par mA, puis par mB - s'écrit v * mA * mB en JavaScript. En toute rigueur syntaxique, cela reviendrait à exiger du développeur JavaScript qu'il précise de gauche à droite une séquence de transformations, c'est-à-dire mB puis mA. Par exemple, pour une translation suivie d'une rotation :

	var m = new libWEBGL.Matrix2D ();
	m.rotate (25);
	m.translate (-0.5, 0.5);

Toutefois, ce serait trop fastidieux. C'est pourquoi l'application de transformations à des matrices a été programmée pour permettre d'écrire simplement :
 
	var m = new libWEBGL.Matrix2D ();
	m.translate (-0.5, 0.5);
	m.rotate (25);

Ces opérations s'appuient sur une méthode de multiplication de matrice m.multiply (m0, m1, ..., mn-1, mn) qui modifie la matrice m de telle sorte qu'un point transformé par elle subit d'abord les transformations de m, puis celles de m0, puis celles de mn, ..., et enfin celles de mn. Autrement dit, m.multiply (m0) revient à une multiplication de ce type (en transposant les matrices pour travailler en ligne, ce qui plus simple à lire pour qui a été formé à travailler avec des matrices et des vecteurs en ligne) :

	                         m0
	     *----------->|  0  1  2  3 |
	     |  *-------->|  4  5  6  7 |
	     |  |  *----->|  8  9 10 11 |
	     |  |  |  *-->| 12 13 14 15 |
         |  |  |  |
	  |  0  1  2  3 | |  0  1  2  3 |
	m |  4  5  6  7 | |  4  5  6  7 |
	  |  8  9 10 11 | |  8  9 10 11 |
	  | 12 13 14 15 | | 12 13 14 15 |

	Par exemple :

	m[0] = m[0] * m0[0] + m[1] * m0[4] + m[2] * m0[8] + m[3] * m0[12]

Une exemple pour le démontrer. Si m est une translation en X, Y et Z et m0 une rotation autour de Z, alors le point transformé par le résultat de m.multiply (m0) subira une translation puis une rotation (toujours en notation en ligne) :

	                         m0
	     *----------->| cos -sin  0 0 |
	     |  *-------->| sin  cos  0 0 |
	     |  |  *----->|   0    0  1 0 |
	     |  |  |  *-->|   0    0  0 1 |
         |  |  |  |
	  |  1  0  0  0 | | cos -sin  0 0 |
	m |  0  1  0  0 | | sin  cos  0 0 |
	  |  0  0  1  0 | |   0    0  1 0 |
	  | dx dy dz  1 | |  rx  ry  dz 1 |

	Avec :
	
	rx = dx * cos + dy * sin
	ry = dx * -sin + dy * cos

	  *------->| cos -sin  0 0 |
	  | *----->| sin  cos  0 0 |
	  | | *--->|   0    0  1 0 |
	  | | | *->|  rx   ry dz 1 |
      | | | |
	| x y z 1 |

	Ce qui produit bien les effets d'une translation suivie d'une rotation :

	x = x * cos + y * sin + rx
	  = (x + dx) * cos + (y + dy) * sin
	y = y * -sin + y * cos + ry
	  = (y + dx) * -sin + (y + dy) * cos
	z = z + dz

Quelques écritures équivalentes pour mA * mB * mC, c'est-à-dire produire la matrice transformant par mA, puis par mB et enfin par mC :

	mA.multiply (mB, mC);

	mA.multiply (mB);
	mA.multiply (mC);

	mC.multiply (mB);
	mA.multiply (mC);

Les coordonnées étant homogènes, des versions optimisées des méthodes de multiplication de matrices, ainsi que des méthodes de multiplication de vecteurs, sont disponibles pour se dispenser d'un certain nombre de multiplications :

	- Matrix3D.multiplyH () considère que la dernière colonne d'une matrice est toujours 0 0 0 1

	- Vector3D.transformH () considère que la dernière composante d'un vecteur est toujours 1

	- Matrix2D.multiplyH () considère que la dernière colonne d'une matrice est toujours 0 0 1

	- Vector2D.transformH () considère que la dernière composante d'un vecteur est toujours 1

Pourquoi ? Pour alléger les calculs effectués dans les applications qui utilisent un unique appel à DrawElements () pour commander le rendu d'un paquet de particules dont chacune doit être transformée par une matrice locale particulière, ce qui impose d'intégrer cette matrice dans le format de point, donc de la calculer dans en JavaScript. Par exemple, l'attribut A_mL dans labFX.html.

Concrèrement, pour reprendre l'exemple donné à l'instant, l'optimisation permet de faire sauter le dernier terme, donc... :

	m[0] = m[0] * m0[0] + m[4] * m0[1] + m[8] * m0[2]

...car :

	m[3] = m0{3] = 0.0
	m[7] = m0[7] = 0.0
	m[11] = m0[11] = 0.0
	m[15] = m0[15] = 1.0

Une conséquence importante de cette optimisation est qu'il ne faut jamais utiliser la matrice de projection ou une matrice inverse dans une telle multiplication, car la dernière colonne ne répond pas aux caractéristiques évoquées. Pour cela, il faut utiliser les versions par défaut :

	- Matrix3D.multiply ()

	- Vector3D.transform ()

	- Matrix2D.multiply ()

	- Vector2D.transform ()

Noter que toutes les méthodes qui s'appuient sur des multiplications de matrices ou des transformations de vecteurs utilisent ces versions par défaut. Par conséquent, elles en sont pas optimisées. Pourquoi ? Parce que l'expérience a démontré qu'utiliser par défaut les versions optimisées dans les méthodes de cette bibliothèque conduit à des bogues compliqués à corriger dans une application qui s'appuient sur cette bibliothèque, si jamais le développeur de l'application oublie que c'est le cas, ceq u'il ne manquera pas, car ce n'est pas intuitif.

Enfin, il faut noter que l'axe des ordonnées est orienté vers le haut dans WebGL, alors qu'il est orienté vers le bas à l'écran. Cela impacte les formules de la matrice de rotation en 2D pour qu'un angle positif corresponde à une rotation dans le sens trigonométrique.
*******************************************************************************/

/*------------------------------------------------------------------------------
Matrice 2D
------------------------------------------------------------------------------*/

libWEBGL.Matrix2D = function (m) {
	this.m = new Array (9);

	if (m === undefined)
		this.identity ();
	else
		this.load (m);
};

libWEBGL.Matrix2D.prototype.load = function (m) {
	var i;

	for (i = 0; i !== 9; i ++)
		this.m[i] = m[i];
};

libWEBGL.Matrix2D.prototype.copy = function (m) {
	this.load (m.m);
};

libWEBGL.Matrix2D.prototype.clone = function () {
	return (new libWEBGL.Matrix2D (this.m));
};

libWEBGL.Matrix2D.prototype.toString = function (precision=3) {
	var j, s;

	s = new Array ();
	for (j = 0; j !== 3; j ++)
		s.push (`[${this.m[j + 3 * 0].toFixed (precision)}\t${this.m[j + 3 * 1].toFixed (precision)}\t${this.m[j + 3 * 2].toFixed (precision)}]`);
	return (s.join ("\n"));
};

libWEBGL.Matrix2D.prototype.identity = function () {
	this.m.fill (0.0);
	this.m[0] = 1.0;
	this.m[4] = 1.0;
	this.m[8] = 1.0;
};

libWEBGL.Matrix2D.prototype.setRotation = function (angle) {
	this.m[0] = Math.cos (angle);
	this.m[3] = -Math.sin (angle);
	this.m[1] = Math.sin (angle);
	this.m[4] = Math.cos (angle);
};

libWEBGL.Matrix2D.prototype.setTranslation = function (dx, dy) {
	this.m[6] = dx;
	this.m[7] = dy;
};

libWEBGL.Matrix2D.prototype.setScaling = function (fx, fy) {
	this.m[0] = fx;
	this.m[4] = fy;
};

libWEBGL.Matrix2D.prototype.rotate = function (angle, isHomogeneous=false) {
	var m;

	m = new libWEBGL.Matrix2D ();
	m.setRotation (angle);
	if (!isHomogeneous)
		this.multiply (m);
	else
		this.multiplyH (m);
};

libWEBGL.Matrix2D.prototype.translate = function (dx, dy, isHomogeneous=false) {
	var m;

	m = new libWEBGL.Matrix2D ();
	m.setTranslation (dx, dy);
	if (!isHomogeneous)
		this.multiply (m);
	else
		this.multiplyH (m);
};

libWEBGL.Matrix2D.prototype.scale = function (fx, fy, isHomogeneous=false) {
	var m;

	m = new libWEBGL.Matrix2D ();
	m.setScaling (fx, fy);
	if (!isHomogeneous)
		this.multiply (m);
	else
		this.multiplyH (m);
};

/*
Pour rappel, le détail de la multiplication des matrices (en ligne pour faciliter la lecture, et non en colonne comme GLSL les traite), en faisant référence aux indices des éléments dans les tableaux JavaScript qui les fournissent colonne après colonne pour alimenter directement GLSL. Donc mA.multiply (mB), si bien qu'un point transformé par le résultat sera transformé par mA puis par mB :

					 mB
		 *------>| 0 1 2 |
		 | *---->| 3 4 5 |
		 | | *-->| 6 7 8 |
		 | | |
	   | 0 1 2 | | 0 1 2 |
	mA | 3 4 5 | | 3 4 5 |
	   | 6 7 8 | | 6 7 8 |

Dans le cas de plusieurs matrices mA.multiply (mB, mC), le point transformé par le résultat sera transformé par mA, puis par mB, puis par mC.
*/

libWEBGL.Matrix2D.prototype.multiply = function (...args) {
	var mA, mB, m, i, j, k;

	mA = this.m;
	for (k = 0; k !== args.length; k ++) {
		mB = args[k];
		m = new Array (9);
		for (j = 0; j !== 3; j ++) {
			for (i = 0; i !== 3; i ++) {
				m[j * 3 + i] = mA[j * 3 + 0] * mB.m[0 * 3 + i] +
							   mA[j * 3 + 1] * mB.m[1 * 3 + i] +
							   mA[j * 3 + 2] * mB.m[2 * 3 + i];
			}
		}
		mA = m;
	}
	this.load (m);
};

// Version optimisée de la multiplication présumant que la dernière colonne des matrices est 0 0 1

libWEBGL.Matrix2D.prototype.multiplyH = function (...args) {
	var mA, mB, m, i, j, k;

	mA = this.m;
	for (k = 0; k !== args.length; k ++) {
		mB = args[k];
		m = new Array (9);
		for (j = 0; j !== 2; j ++) {
			for (i = 0; i !== 2; i ++) {
				m[j * 3 + i] = mA[j * 3 + 0] * mB.m[0 * 3 + i] +
							   mA[j * 3 + 1] * mB.m[1 * 3 + i];
			}
		}
		for (i = 0; i !== 2; i ++) {
			m[j * 3 + i] = mA[j * 3 + 0] * mB.m[0 * 3 + i] +
						   mA[j * 3 + 1] * mB.m[1 * 3 + i] +
						   mA[j * 3 + 2] * mB.m[2 * 3 + i];
		}
		m[2] = 0.0;
		m[5] = 0.0;
		m[8] = 1.0;
		mA = m;
	}
	this.load (m);
};

libWEBGL.Matrix2D.prototype.inverse = function () {
	var c0, c1, c2, d, m;

	c2 = this.m[3] * this.m[7] - this.m[4] * this.m[6];
	c1 = this.m[3] * this.m[8] - this.m[5] * this.m[6];
	c0 = this.m[4] * this.m[8] - this.m[5] * this.m[7];
	d = (this.m[0] * c0 - this.m[1] * c1 + this.m[2] * c2);

	m = new Array (9);
	m[0] = c0 / d;
	m[3] = - c1 / d;
	m[6] = c2 / d;
	m[1] = - (this.m[1] * this.m[8] - this.m[7] * this.m[2]) / d;
	m[4] = (this.m[0] * this.m[8] - this.m[6] * this.m[2]) / d;
	m[7] = - (this.m[0] * this.m[7] - this.m[6] * this.m[1]) / d;
	m[2] = (this.m[1] * this.m[5] - this.m[4] * this.m[2]) / d;
	m[5] = - (this.m[0] * this.m[5] - this.m[3] * this.m[2]) / d;
	m[8] = (this.m[0] * this.m[4] - this.m[3] * this.m[1]) / d;
	this.load (m);
};

/*------------------------------------------------------------------------------
Vecteur 2D
------------------------------------------------------------------------------*/

libWEBGL.Vector2D = function (v) {
	this.v = new Array (3);

	if (v === undefined)
		this.origin ();
	else
		this.load (v);
};

Object.defineProperty (libWEBGL.Vector2D.prototype, "x", {
	get: function () { return (this.v[0]); },
	set: function (x) { this.v[0] = x; }
});

Object.defineProperty (libWEBGL.Vector2D.prototype, "y", {
	get: function () { return (this.v[1]); },
	set: function (y) { this.v[1] = y; }
});

Object.defineProperty (libWEBGL.Vector2D.prototype, "w", {
	get: function () { return (this.v[2]); },
	set: function (w) { this.v[2] = w; }
});

libWEBGL.Vector2D.prototype.load = function (v) {
	var i;

	for (i = 0; i !== 3; i ++)
		this.v[i] = v[i];
	return (this);
};

libWEBGL.Vector2D.prototype.copy = function (v) {
	this.load (v.v);
	return (this);
};

libWEBGL.Vector2D.prototype.clone = function () {
	return (new libWEBGL.Vector2D (this.v));
};

libWEBGL.Vector2D.prototype.toString = function (precision=3) {
	s = `(${this.v[0].toFixed (precision)}, ${this.v[1].toFixed (precision)}, ${this.v[2].toFixed (precision)})`;
	return (s);
};

libWEBGL.Vector2D.prototype.origin = function () {
	this.v[0] = 0.0;
	this.v[1] = 0.0;
	this.v[2] = 1.0;
	return (this);
};

libWEBGL.Vector2D.prototype.scale = function (k) {
	this.v[0] *= k;
	this.v[1] *= k;
	return (this);
};

libWEBGL.Vector2D.prototype.sub = function (v) {
	this.v[0] -= v.v[0];
	this.v[1] -= v.v[1];
	return (this);
};

libWEBGL.Vector2D.prototype.add = function (v) {
	this.v[0] += v.v[0];
	this.v[1] += v.v[1];
	return (this);
};

/*
Pour rappel, le détail de la multiplication d'un vecteur par une matrice (en ligne pour faciliter la lecture, et non en colonne comme GLSL les traite), en faisant référence aux indices des éléments dans les tableaux JavaScript qui les fournissent colonne après colonne pour alimenter directement GLSL. Donc v.multiply (m) :

	                 m
		 *------>| 0 1 2 |
		 | *---->| 3 4 5 |
		 | | *-->| 6 7 8 |
		 | | |
	 v | 0 1 2 | | 0 1 2 |

Dans le cas de plusieurs matrices v.transform (mA, mB), le point est transformé par mA, puis par mB.
*/

libWEBGL.Vector2D.prototype.transform = function (...args) {
	var w, m, v, i;

	w = this.v;
	for (m of args) {
		v = new Array (3);
		for (i = 0; i !== 3; i ++) {
			v[i] = w[0] * m.m[0 * 3 + i] +
				   w[1] * m.m[1 * 3 + i] +
				   w[2] * m.m[2 * 3 + i];
		}
		w = v;
	}
	this.load (v);
	return (this);
};

// Version optimisée de la transformation présumant que la composante w vaut 1, et que la dernière colonne de la matrice est 0 0 1

libWEBGL.Vector2D.prototype.transformH = function (...args) {
	var w, m, v, i;

	w = this.v;
	for (m of args) {
		v = new Array (3);
		for (i = 0; i !== 2; i ++) {
			v[i] = this.v[0] * m.m[0 * 3 + i] +
				   this.v[1] * m.m[1 * 3 + i] +
				   m.m[2 * 3 + i];
		}
		w = v;
	}
	v[2] = 1.0;
	this.load (v);
	return (this);
};

libWEBGL.Vector2D.prototype.normalize = function (v) {
	var d;

	d = Math.sqrt (this.v[0] * this.v[0] + this.v[1] * this.v[1]);
	this.v[0] /= d;
	this.v[1] /= d;
	return (this);
};

libWEBGL.Vector2D.prototype.dot = function (v) {
	var p;

	p = this.v[0] * v.v[0] + this.v[1] * v.v[1];
	return (p);
};

libWEBGL.Vector2D.prototype.cross = function (v) {
	var d;

	d = this.v[0] * v.v[1] - this.v[1] * v.v[0];
	return (d);
};

/*------------------------------------------------------------------------------
Matrice 3D
------------------------------------------------------------------------------*/

libWEBGL.Matrix3D = function (m) {
	this.m = new Array (16);

	if (m === undefined)
		this.identity ();
	else
		this.load (m);
};

libWEBGL.Matrix3D.prototype.load = function (m) {
	var i;

	for (i = 0; i !== 16; i ++)
		this.m[i] = m[i];
	return (this);
};

libWEBGL.Matrix3D.prototype.copy = function (m) {
	this.load (m.m);
	return (this);
};

libWEBGL.Matrix3D.prototype.clone = function () {
	return (new libWEBGL.Matrix3D (this.m));
};

libWEBGL.Matrix3D.prototype.toString = function (precision=3) {
	var j, s;

	s = new Array ();
	for (j = 0; j !== 4; j ++)
		s.push (`[${this.m[j + 3 * 0].toFixed (precision)}\t${this.m[j + 3 * 1].toFixed (precision)}\t${this.m[j + 3 * 2].toFixed (precision)}\t${this.m[j + 3 * 3].toFixed (precision)}]`);
	return (s.join ("\n"));
};

libWEBGL.Matrix3D.prototype.identity = function () {
	this.m.fill (0.0);
	this.m[0] = 1.0;
	this.m[5] = 1.0;
	this.m[10] = 1.0;
	this.m[15] = 1.0;
	return (this);
};

libWEBGL.Matrix3D.prototype.setRotationX = function (angle) {
	this.m[5] = Math.cos (angle);
	this.m[6] = Math.sin (angle);
	this.m[9] = -Math.sin (angle);
	this.m[10] = Math.cos (angle);
	return (this);
};

libWEBGL.Matrix3D.prototype.setRotationY = function (angle) {
	this.m[0] = Math.cos (angle);
	this.m[2] = -Math.sin (angle);
	this.m[8] = Math.sin (angle);
	this.m[10] = Math.cos (angle);
	return (this);
};

libWEBGL.Matrix3D.prototype.setRotationZ = function (angle) {
	this.m[0] = Math.cos (angle);
	this.m[1] = Math.sin (angle);
	this.m[4] = -Math.sin (angle);
	this.m[5] = Math.cos (angle);
	return (this);
};

libWEBGL.Matrix3D.prototype.setTranslation = function (dx, dy, dz) {
	this.m[12] = dx;
	this.m[13] = dy;
	this.m[14] = dz;
	return (this);
};

libWEBGL.Matrix3D.prototype.setScaling = function (fx, fy, fz) {
	this.m[0] = fx;
	this.m[5] = fy;
	this.m[10] = fz;
	return (this);
};

libWEBGL.Matrix3D.prototype.rotateX = function (angle, isHomogeneous=false) {
	var m;

	m = new libWEBGL.Matrix3D ();
	m.setRotationX (angle);
	if (!isHomogeneous)
		this.multiply (m);
	else
		this.multiplyH (m);
	return (this);
};

libWEBGL.Matrix3D.prototype.rotateY = function (angle, isHomogeneous=false) {
	var m;

	m = new libWEBGL.Matrix3D ();
	m.setRotationY (angle);
	if (!isHomogeneous)
		this.multiply (m);
	else
		this.multiplyH (m);
	return (this);
};

libWEBGL.Matrix3D.prototype.rotateZ = function (angle, isHomogeneous=false) {
	var m;

	m = new libWEBGL.Matrix3D ();
	m.setRotationZ (angle);
	if (!isHomogeneous)
		this.multiply (m);
	else
		this.multiplyH (m);
	return (this);
};

libWEBGL.Matrix3D.prototype.translate = function (dx, dy, dz, isHomogeneous=false) {
	var m;

	m = new libWEBGL.Matrix3D ();
	m.setTranslation (dx, dy, dz);
	if (!isHomogeneous)
		this.multiply (m);
	else
		this.multiplyH (m);
	return (this);
};

libWEBGL.Matrix3D.prototype.scale = function (fx, fy, fz, isHomogeneous=false) {
	var m;

	m = new libWEBGL.Matrix3D ();
	m.setScaling (fx, fy, fz);
	if (!isHomogeneous)
		this.multiply (m);
	else
		this.multiplyH (m);
	return (this);
};

/*
Pour rappel, le détail multiplication des matrices (en ligne pour faciliter la lecture, et non en colonne comme GLSL les traite), en faisant référence aux indices des éléments dans les tableaux JavaScript qui les fournissent colonne après colonne pour alimenter directement GLSL. Donc mA.multiply (mB), si bien qu'un point transformé par le résultat sera transformé par mA puis par mB :

	                          mB
	      *----------->|  0  1  2  3 |
	      |  *-------->|  4  5  6  7 |
	      |  |  *----->|  8  9 10 11 |
	      |  |  |  *-->| 12 13 14 15 |
          |  |  |  |
	   |  0  1  2  3 | |  0  1  2  3 |
	mA |  4  5  6  7 | |  4  5  6  7 |
	   |  8  9 10 11 | |  8  9 10 11 |
	   | 12 13 14 15 | | 12 13 14 15 |

Dans le cas de plusieurs matrices mA.multiply (mB, mC), le point transformé par le résultat sera transformé par mA, puis par mB, puis par mC.
*/

libWEBGL.Matrix3D.prototype.multiply = function (...args) {
	var mA, mB, m, i, j, k;

	mA = this.m;
	for (k = 0; k !== args.length; k ++) {
		mB = args[k];
		m = new Array (16);
		for (j = 0; j !== 4; j ++) {
			for (i = 0; i !== 4; i ++) {
				m[j * 4 + i] = mA[j * 4 + 0] * mB.m[0 * 4 + i] +
							   mA[j * 4 + 1] * mB.m[1 * 4 + i] +
							   mA[j * 4 + 2] * mB.m[2 * 4 + i] +
							   mA[j * 4 + 3] * mB.m[3 * 4 + i];
			}
		}
		mA = m;
	}
	this.load (m);
	return (this);
};

// Version optimisée de la multiplication présumant que la dernière colonne des matrices est 0 0 0 1

libWEBGL.Matrix3D.prototype.multiplyH = function (...args) {
	var mA, mB, m, i, j, k;

	mA = this.m;
	for (k = 0; k !== args.length; k ++) {
		mB = args[k];
		m = new Array (16);
		for (j = 0; j !== 3; j ++) {
			for (i = 0; i !== 3; i ++) {
				m[j * 4 + i] = mA[j * 4 + 0] * mB.m[0 * 4 + i] +
							   mA[j * 4 + 1] * mB.m[1 * 4 + i] +
							   mA[j * 4 + 2] * mB.m[2 * 4 + i];
			}
		}
		for (i = 0; i !== 3; i ++) {
			m[j * 4 + i] = mA[j * 4 + 0] * mB.m[0 * 4 + i] +
						   mA[j * 4 + 1] * mB.m[1 * 4 + i] +
						   mA[j * 4 + 2] * mB.m[2 * 4 + i] +
						   mA[j * 4 + 3] * mB.m[3 * 4 + i];
		}
		m[3] = 0.0;
		m[7] = 0.0;
		m[11] = 0.0;
		m[15] = 1.0;
		mA = m;
	}
	this.load (m);
	return (this);
};

libWEBGL.Matrix3D.prototype.inverse = function (m) {
	var s0, s1, s2, s3, s4, s5, c0, c1, c2, c3, c4, c5, d, m;

	s0 = this.m[0] * this.m[5] - this.m[4] * this.m[1];
	s1 = this.m[0] * this.m[6] - this.m[4] * this.m[2];
	s2 = this.m[0] * this.m[7] - this.m[4] * this.m[3];
	s3 = this.m[1] * this.m[6] - this.m[5] * this.m[2];
	s4 = this.m[1] * this.m[7] - this.m[5] * this.m[3];
	s5 = this.m[2] * this.m[7] - this.m[6] * this.m[3];
	c5 = this.m[10] * this.m[15] - this.m[14] * this.m[11];
	c4 = this.m[9] * this.m[15] - this.m[13] * this.m[11];
	c3 = this.m[9] * this.m[14] - this.m[13] * this.m[10];
	c2 = this.m[8] * this.m[15] - this.m[12] * this.m[11];
	c1 = this.m[8] * this.m[14] - this.m[12] * this.m[10];
	c0 = this.m[8] * this.m[13] - this.m[12] * this.m[9];
	d = (s0 * c5 - s1 * c4 + s2 * c3 + s3 * c2 - s4 * c1 + s5 * c0);

	m = new Array (16);
	m[0] = (this.m[5] * c5 - this.m[6] * c4 + this.m[7] * c3) / d;
	m[1] = (-this.m[1] * c5 + this.m[2] * c4 - this.m[3] * c3) / d;
	m[2] = (this.m[13] * s5 - this.m[14] * s4 + this.m[15] * s3) / d;
	m[3] = (-this.m[9] * s5 + this.m[10] * s4 - this.m[11] * s3) / d;
	m[4] = (-this.m[4] * c5 + this.m[6] * c2 - this.m[7] * c1) / d;
	m[5] = (this.m[0] * c5 - this.m[2] * c2 + this.m[3] * c1) / d;
	m[6] = (-this.m[12] * s5 + this.m[14] * s2 - this.m[15] * s1) / d;
	m[7] = (this.m[8] * s5 - this.m[10] * s2 + this.m[11] * s1) / d;
	m[8] = (this.m[4] * c4 - this.m[5] * c2 + this.m[7] * c0) / d;
	m[9] = (-this.m[0] * c4 + this.m[1] * c2 - this.m[3] * c0) / d;
	m[10] = (this.m[12] * s4 - this.m[13] * s2 + this.m[15] * s0) / d;
	m[11] = (-this.m[8] * s4 + this.m[9] * s2 - this.m[11] * s0) / d;
	m[12] = (-this.m[4] * c3 + this.m[5] * c1 - this.m[6] * c0) / d;
	m[13] = (this.m[0] * c3 - this.m[1] * c1 + this.m[2] * c0) / d;
	m[14] = (-this.m[12] * s3 + this.m[13] * s1 - this.m[14] * s0) / d;
	m[15] = (this.m[8] * s3 - this.m[9] * s1 + this.m[10] * s0) / d;
	this.load (m);
	return (this);
};

libWEBGL.Matrix3D.prototype.projection = function (fov, near, far, width, height) {
	var left, right, top, bottom, aspectRatioX, aspectRatioY, m;

	if (width > height) {
		aspectRatioX = width / height;
		aspectRatioY = 1.0;
	}
	else {
		aspectRatioX = 1.0;
		aspectRatioY = height / width;
	}
	right = aspectRatioX * -near * Math.tan (fov / 2.0);
	left = -right;
	top = aspectRatioY * -near * Math.tan (fov / 2.0);
	bottom = -top;
	m = [
		-2.0 * near / (right - left), 0.0, 0.0, 0.0,
		0.0, -2.0 * near / (top - bottom), 0.0, 0.0,
		(left + right) / (right - left), (bottom + top) / (top - bottom), (near + far) / (near - far), -1.0,
		0.0, 0.0, 2.0 * near * far / (far - near), 0.0
	];
	this.load (m);
	return (this);
};

/*------------------------------------------------------------------------------
Vecteur 3D
------------------------------------------------------------------------------*/

libWEBGL.Vector3D = function (v) {
	this.v = new Array (4);

	if (v === undefined)
		this.origin ();
	else
		this.load (v);
};

Object.defineProperty (libWEBGL.Vector3D.prototype, "x", {
	get: function () { return (this.v[0]); },
	set: function (x) { this.v[0] = x; }
});

Object.defineProperty (libWEBGL.Vector3D.prototype, "y", {
	get: function () { return (this.v[1]); },
	set: function (y) { this.v[1] = y; }
});

Object.defineProperty (libWEBGL.Vector3D.prototype, "z", {
	get: function () { return (this.v[2]); },
	set: function (z) { this.v[2] = z; }
});

Object.defineProperty (libWEBGL.Vector3D.prototype, "w", {
	get: function () { return (this.v[3]); },
	set: function (w) { this.v[3] = w; }
});

libWEBGL.Vector3D.prototype.load = function (v) {
	var i;

	for (i = 0; i !== 4; i ++)
		this.v[i] = v[i];
	return (this);
};

libWEBGL.Vector3D.prototype.copy = function (v) {
	this.load (v.v);
	return (this);
};

libWEBGL.Vector3D.prototype.clone = function () {
	return (new libWEBGL.Vector3D (this.v));
};

libWEBGL.Vector3D.prototype.toString = function (precision=3) {
	s = `(${this.v[0].toFixed (precision)}, ${this.v[1].toFixed (precision)}, ${this.v[2].toFixed (precision)}, ${this.v[3].toFixed (precision)})`;
	return (s);
};

libWEBGL.Vector3D.prototype.origin = function () {
	this.v[0] = 0.0;
	this.v[1] = 0.0;
	this.v[2] = 0.0;
	this.v[3] = 1.0;
	return (this);
};

libWEBGL.Vector3D.prototype.scale = function (k) {
	this.v[0] *= k;
	this.v[1] *= k;
	this.v[2] *= k;
	return (this);
};

libWEBGL.Vector3D.prototype.sub = function (v) {
	this.v[0] -= v.v[0];
	this.v[1] -= v.v[1];
	this.v[2] -= v.v[2];
	return (this);
};

libWEBGL.Vector3D.prototype.add = function (v) {
	this.v[0] += v.v[0];
	this.v[1] += v.v[1];
	this.v[2] += v.v[2];
	return (this);
};

/*
Pour rappel, le détail de la multiplication d'un vecteur par une matrice (en ligne pour faciliter la lecture, et non en colonne comme GLSL les traite), en faisant référence aux indices des éléments dans les tableaux JavaScript qui les fournissent colonne après colonne pour alimenter directement GLSL. Donc v.multiply (m) :

	                      m
		 *-------->|  0  1  2  3 |
		 | *------>|  4  5  6  7 |
		 | | *---->|  8  9 10 11 |
		 | | | *-->| 12 13 14 15 |
		 | | | |
	 v | 0 1 2 3 | |  0  1  2  3 |

Dans le cas de plusieurs matrices v.transform (mA, mB), le point est transformé par mA, puis par mB.
*/

libWEBGL.Vector3D.prototype.transform = function (...args) {
	var w, m, v, i;

	w = this.v;
	for (m of args) {
		v = new Array (4);
		for (i = 0; i !== 4; i ++) {
			v[i] = w[0] * m.m[0 * 4 + i] +
				   w[1] * m.m[1 * 4 + i] +
				   w[2] * m.m[2 * 4 + i] +
				   w[3] * m.m[3 * 4 + i];
		}
		w = v;
	}
	this.load (v);
	return (this);
};

// Version optimisée de la transformation présumant que la composante w vaut 1, et que la dernière colonne de la matrice est 0 0 0 1

libWEBGL.Vector3D.prototype.transformH = function (...args) {
	var w, m, v, i;

	w = this.v;
	for (m of args) {
		v = new Array (4);
		for (i = 0; i !== 3; i ++) {
			v[i] = w[0] * m.m[0 * 4 + i] +
				   w[1] * m.m[1 * 4 + i] +
				   w[2] * m.m[2 * 4 + i] +
				   m.m[3 * 4 + i];
		}
		w = v;
	}
	v[3] = 1.0;
	this.load (v);
	return (this);
};

libWEBGL.Vector3D.prototype.normalize = function (v) {
	var d;

	d = Math.sqrt (this.v[0] * this.v[0] + this.v[1] * this.v[1] + this.v[2] * this.v[2]);
	this.v[0] /= d;
	this.v[1] /= d;
	this.v[2] /= d;
	return (this);
};

libWEBGL.Vector3D.prototype.dot = function (v) {
	var p;

	p = this.v[0] * v.v[0] + this.v[1] * v.v[1] + this.v[2] * v.v[2];
	return (p);
};

libWEBGL.Vector3D.prototype.cross = function (v) {
	var n;

	n = [
		this.v[1] * v.v[2] - this.v[2] * v.v[1],
		this.v[2] * v.v[0] - this.v[0] * v.v[2],
		this.v[0] * v.v[1] - this.v[1] * v.v[0],
		1.0
	];
	return (libWEBGL.createVector3D (n));
};
